#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The `reducing` utilities provides methods to reduce (dimensions) of arrays.

@author: hoelken
"""
from typing import Iterable


def flatten_sorted(lists: Iterable) -> list:
    """Flattens a list of lists and sorts the result"""
    result = flatten(lists)
    result.sort()
    return result


def flatten(lists: Iterable) -> list:
    """Flattens a list of lists and sorts the result"""
    result = []
    for entry in lists:
        if isinstance(entry, Iterable) and not isinstance(entry, (str, dict)):
            result.extend(flatten(entry))
        else:
            result.append(entry)
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
provides housekeeping / setup methods to reduce the programming overhead of spawning threads or processes.
"""

import os
import subprocess
import time
import logging
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from functools import partial

from .thread import Thread


log = logging.getLogger()


CPU_LIM = round(os.cpu_count() * 0.7)
"""70% of CPUs from the current machine (max usage limit)"""


def threaded(func, args, workers=10, raise_exception=True):
    """
    Calls the given function in multiple threads for the set of given arguments
        Note that this does not spawn processes, but threads. Use this for non CPU
        CPU dependent tasks, i.e. I/O
    Method returns once all calls are done.

    ### Params
    - func: [Function] the function to call
    - args: [Iterable] the 'list' of arguments for each call
    - workers: [Integer] the number of concurrent threads to use
    - raise_exception: [Bool] Flag if an exception in a thread shall be raised or just logged

    ### Returns
    Results from all `Threads` as list
    """
    if len(args) == 1:
        return list(func(arg) for arg in args)

    subprocess.Popen.communicate = partial(subprocess.Popen.communicate, timeout=30)
    with ThreadPoolExecutor(workers) as ex:
        threads = [Thread(func, arg).submit(ex) for arg in args]
    return _collect_results(threads, raise_exception)


def simultaneous(func, args, workers: int = None, raise_exception: bool = True):
    """
    Calls the given function in multiple processes for the set of given arguments
        Note that this does spawn processes, not threads. Use this for task that
        depend heavily on CPU and can be done in parallel.
    Method returns once all calls are done.

    ### Params
    - func: [Function] the function to call
    - args: [Iterable] the 'list' of arguments for each call
    - workers: [Integer] the number of concurrent threads to use (Default: NUM_CPUs)
    - raise_exception: [Bool] Flag if an exception in a thread shall be raised or just logged

    ### Returns
    Results from all `Threads` as list
    """
    if len(args) == 1:
        return list(func(arg) for arg in args)

    if workers is None:
        workers = CPU_LIM
    with ProcessPoolExecutor(workers) as ex:
        threads = [Thread(func, arg).submit(ex) for arg in args]
    return _collect_results(threads, raise_exception)


def _collect_results(threads: list, raise_exception: bool = True) -> list:
    """
    Takes a list of threads and busy waits for them to be executed.

    ### Params
    - threads: [List<Thread>] a list of submitted threads
    - raise_exception: [Bool] Flag if an exception in a thread shall be raised or just logged

    ### Returns
    Results from all `Threads` as list
    """
    result = []
    while len(threads) > 0:
        for thread in threads:
            if not thread.is_submitted():
                log.debug('Removing not submitted thread.')
                threads.remove(thread)
            if not thread.is_done():
                continue

            if thread.exception is not None:
                _exception_handling(threads, thread, raise_exception)
            else:
                result.append(thread.result)
            threads.remove(thread)
        if len(threads):
            time.sleep(0.1)
    return result


def _exception_handling(threads, thread, raise_exception):
    ex = thread.exception
    print('')
    log.critical("Execution of '%s' caused\n[%s]: %s",
                 thread.function.__name__, ex.__class__.__name__, ex)
    if raise_exception:
        # Stop all remaining threads:
        for t in threads:
            t.cancel()
        raise ex



from typing import Union


def parse_shape(string: str) -> tuple:
    """
    Parses a numpy like shape string, e.g. [0:,42:1337]
    :returns: a tuple of slices that can be used with numpy arrays
    """
    string = string.strip("[]")
    return tuple(to_slice(d) if ':' in d else to_int(d) for d in string.split(","))


def to_slice(string: str, delimiter: str = ':') -> slice:
    """:returns: a slice with start and stop taken from the string, seperated by delimiter, e.g. 3:7"""
    start, stop = string.split(delimiter)
    return slice(to_int(start), to_int(stop))


def to_int(string: str) -> Union[int, None]:
    """:returns: an integer if the string is numeric"""
    if string.strip().isnumeric():
        return int(string)
    return None


def shape_slices_to_string(shape: tuple) -> str:
    out = []
    for dim in shape:
        if isinstance(dim, slice):
            out.append(f'{dim.start}:{dim.stop}')
        else:
            out.append(str(dim))
    return f"[{', '.join(out)}]"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Mocks for testing multi processing functionality.

@author: hoelken
"""


def sequentially(func, args, workers=None):
    """
    Method mimics the `simultaneous` or `threaded` methods, but calls everything in sequence.
    :param func: function to call
    :param args: arguments to call the function with
    :param workers: Ignored
    :return: result of the call.
    """
    return list(func(arg) for arg in args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The threads hold information on the function (and result) to execute in a thread or process.

@author: hoelken
"""

import logging

from concurrent.futures import Executor

log = logging.getLogger()


class Thread:
    """
    # Thread
    The threads hold information on the function to execute in a thread or process.
    Provides an interface to the `future` object once submitted to an executer.
    """

    def __init__(self, func: callable, args: object):
        self.function = func
        self.arguments = args
        self.future = None

    def submit(self, executor: Executor):
        """Start execution via executor"""
        if not self.is_submitted():
            self.future = executor.submit(self.function, self.arguments)
        return self

    def is_submitted(self) -> bool:
        return self.future is not None

    def is_done(self) -> bool:
        return self.is_submitted() and self.future.done()

    @property
    def exception(self):
        if not self.is_done():
            return None
        return self.future.exception()

    @property
    def result(self):
        if not self.is_submitted():
            return None
        return self.future.result()

    def cancel(self):
        try:
            self.future.cancel()
        except RuntimeError as e:
            log.warning('Unable to cancel thread: %s', e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
module provides functions to print progress to the shell (and only the shell)
"""

RESET = '\033[1;0m'
GREEN = '\033[1;32m'
RED = '\033[1;31m'


def dot(success: bool = True, flush: bool = False, char: str = '*') -> None:
    """Print a progress dot to std. out"""
    if flush:
        print('')
    else:
        print(f'{GREEN if success else RED}{char}{RESET}', end='', flush=True)


def msg(txt: str = '', flush: bool = False) -> None:
    if flush:
        print('')
    else:
        print(txt, end='\r', flush=True)


def bar(current: float, total: float, flush: bool = False, length: int = 50) -> None:
    """print a progress bar and a percent of completion to std. out """
    percent = (current * 100) / total
    done = int(round(length * (percent/100)))
    prog = ''.join(['━' for _ in range(done)])
    empty = ''.join([' ' for _ in range(length - done)])
    stop = '┃' if len(empty) > 0 else '┫'
    print(f'┣{prog}{empty}{stop} \t[{percent:.2f} %]', end='\r', flush=True)
    if flush:
        print('')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The `chunking` utility provides methods to break huge collections in chunks

@author: hoelken
"""


def chunker(seq: list, size: int) -> list:
    """
    Generates chunks (slices) from a given sequence

    ### Params
    - seq: the list to chunk
    - size: the size of the chunks

    ### Returns
    A list of lists where each list has the
        length of the requested chunk size (maybe except the last one)
    """
    if size < 1:
        return [seq]
    return [seq[pos:pos + size] for pos in range(0, len(seq), size)]


def indexed_chunks(seq: list, size: int) -> dict:
    """
    Generates indexed chunks (slices) from a given sequence
    ### Params
    - seq: List the list to chunk
    - size: Integer the size of the chunks

    ### Returns
     A dictionary with the index as key and the corresponding chunk as value.
        The length of the value arrays is the requested chunk size (maybe except the last one)
    """
    idx = 0
    indexed = {}
    for chunk in chunker(seq, size):
        indexed[idx] = chunk
        idx += 1
    return indexed
